﻿using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;
using System;

namespace NativeFileDialogSharp.Native;

internal unsafe static class LibResolver
{
    public const string LibName = NativeFunctions.LibraryName;

    internal static nint? CachedPtr;

    static LibResolver()
    {
        NativeLibrary.SetDllImportResolver(Assembly.GetExecutingAssembly(), OnDllImport);
    }

    public static void EnsureInit() { }

    private static nint OnDllImport(string libraryName, Assembly assembly, DllImportSearchPath? searchPath)
    {
        if (libraryName == NativeFunctions.LibraryName)
        {
            if (CachedPtr is nint cached) return cached;
            if (TryFindLib(libraryName, assembly, searchPath, out var lib))
            {
                CachedPtr = lib;
                return lib;
            }
        }

        return NativeLibrary.Load(libraryName, assembly, searchPath);
    }

    private static bool TryFindLib(string name, Assembly assembly, DllImportSearchPath? searchPath, out nint lib)
    {
        string rid = RuntimeInformation.RuntimeIdentifier;
        string nugetNativeLibsPath = Path.Combine(AppContext.BaseDirectory, "runtimes", rid, "native");
        bool isNuGetRuntimeLibrariesDirectoryPresent = Directory.Exists(nugetNativeLibsPath);
        string dllName = name;
        string altDllName = null;

        if (OperatingSystem.IsWindows())
        {
            dllName = $"{name}.dll";

            if (!isNuGetRuntimeLibrariesDirectoryPresent)
            {
                rid = RuntimeInformation.ProcessArchitecture switch
                {
                    Architecture.X64 => "win-x64",
                    Architecture.X86 => "win-x86",
                    _ => throw new NotSupportedException()
                };

                nugetNativeLibsPath = Path.Combine(AppContext.BaseDirectory, "runtimes", rid, "native");
                isNuGetRuntimeLibrariesDirectoryPresent = Directory.Exists(nugetNativeLibsPath);
            }
        }
        else if (OperatingSystem.IsLinux())
        {
            dllName = $"{name}.so";
            altDllName = $"lib{name}.so";
        }

        if (isNuGetRuntimeLibrariesDirectoryPresent)
        {
            string nativeLibPath = Path.Combine(AppContext.BaseDirectory, "runtimes", rid, "native", dllName);

            if (NativeLibrary.TryLoad(nativeLibPath, out lib))
            {
                return true;
            }

            if (altDllName != null)
            {
                nativeLibPath = Path.Combine(AppContext.BaseDirectory, "runtimes", rid, "native", altDllName);

                if (NativeLibrary.TryLoad(nativeLibPath, out lib))
                {
                    return true;
                }
            }
        }
        else
        {
            if (NativeLibrary.TryLoad(dllName, assembly, searchPath, out lib))
            {
                return true;
            }
        }

        lib = IntPtr.Zero;
        return false;
    }
}
